/* landscapes.h */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DYNAPRO implements a dynamic programming algorithm for predicting
DNA occupancies by arbitrary objects whose interaction energies with DNA are
known for each possible binding position.

The DYNAPRO method and its applications are described in the following
paper:

******************************************************************/

#ifndef _landscapes_h_
#define _landscapes_h_

#include "sequence.h"
#include "stats.h"

#define NUC_NUM 4

std::vector<std::string> ReadList(std::string const &filename);
void ReadList_double(std::string const &filename, std::vector<std::string> &c1, std::vector<int> &c2);

// This class contains background energies for each of the four nucleotides.
// Currently they depend only on the type of the nuc (in the {A,C,G,T} order)
// but can be easily extended to any shape as long as it's fixed on the seq.
class Bkgr {
 private:
  double Eb[NUC_NUM];
 public:
  Bkgr();
  Bkgr(std::string const &bkgr_choice);
  ~Bkgr() {};
  double GetEb(unsigned int ind);
  double GetEb(char nuc);
};

// This class contains values of Pmin (min allowed posterior probs).
// They are either provided by the user or set to -1.0 so that
// everything is allowed.
class Probmin {
 private:
  double *pmin;
  unsigned int Lpmin; // how many Pmins are stored
 public:
  friend std::vector<std::string> ReadList(std::string const &filename);
  Probmin();
  Probmin(std::string const &Pmin_choice, unsigned int L);
  ~Probmin();
  double GetPmin(unsigned int ind);
  unsigned int GetLPmin();
};

// This class contains energy landscapes used as input to the dynamic programming routine.
class Landscapes {
 private:
  std::vector< std::vector<double> > en; // 2D array of input energies
  std::vector< std::vector<double> > en_rc; // 2D array of revcom input energies
  std::vector< std::vector<int> > seqN; // 2D array of seqnums
  std::vector<double> meanEs; // corresponding <E> and var(E)
  std::vector<double> varEs;
  std::vector<int> Lembed; // values of Lembed
  std::vector<int> Ls; // footprints (in bp) for all input objects, deduced from genomic seq and energy landscapes
  std::vector< std::vector<double> > prob;
  std::vector< std::vector<double> > occ;
  unsigned int NLAND; // total number of landscapes (incl. bkgr)
  void build_bkgr(Bkgr &bkgr, Sequence *seq);
  void read_energy_file_processed(std::string const &etname, unsigned int ind);
  void combine_strands(std::vector<double> &E, std::vector<double> &Erc, double Bw);
  void rescale_histogram(std::vector<double> &E, double mean, double var);
  void get_footprints(unsigned int Lseq);
  void embed_landscape(Bkgr &bkgr, Sequence *seq, unsigned int ind);
 public:
  friend void ReadList_double(std::string const &filename, std::vector<std::string> &c1, std::vector<int> &c2);
  Landscapes();
  Landscapes(std::string const &filename, double Bw, Bkgr &bkgr, Sequence *seq);
  ~Landscapes();
  unsigned int GetNLAND();
  void dynapro(int Lseq, Probmin &Pmin);
  void output_prob_occ(std::string const &out, std::string const &bkgr_choice, unsigned int Lseq, double Bw);
};

#endif

