#!/usr/bin/perl -w

use strict;

# This script is used to generate an energy landscape for a single factor
# by computing PWM log-scores: \sum_i^m log(w_i^{\alpha_i}/b^{\alpha_i}).
# There is a companion script: generate_factor_landscape.pl which
# generates the inverted tophat shape potential.

my $PSEUDO = 1e-02;

#my @Pb = (0.25, 0.25, 0.25, 0.25);
my @Pb = (0.309012, 0.191674, 0.191303, 0.308012);

#my $Bw = 0.5; # strand bias

my %roman2arabic = ('I' => 1, 'II' => 2, 'III' => 3, 'IV' => 4,
		    'V' => 5, 'VI' => 6, 'VII' => 7, 'VIII' => 8,
		    'IX' => 9, 'X' => 10, 'XI' => 11, 'XII' => 12,
		    'XIII' => 13, 'XIV' => 14, 'XV' => 15, 'XVI' => 16,
		    '1' => 1, '2' => 2, '3' => 3, '4' => 4,
		    '5' => 5, '6' => 6, '7' => 7, '8' => 8,
		    '9' => 9, '10' => 10, '11' => 11, '12' => 12,
		    '13' => 13, '14' => 14, '15' => 15, '16' => 16);

my %base_map = ('A' => 0, 'C' => 1, 'G' => 2, 'T' => 3);
my %base_rc = ('A' => 'T', 'C' => 'G', 'G' => 'C', 'T' => 'A');

# argv
my %opts = &getCommandLineOptions();
my $fasta    = $opts{f}; # FASTA file with genomic sequence
my $pwm_file = $opts{pwm}; # PWM using which the landscape will be generated
my $gff      = $opts{gff}; # optional GFF file with site coords (everything else will be treated as background & made unfavorable)
my $Emean    = $opts{mean}; # constant bkgr energy for the whole landscape ("bkgr E")
######## GFF file filters: #########
my $chrN     = $opts{chrN}; # chr_num - keep only sites at this chr from the GFF file
my $TFname   = $opts{TF}; # name of the TF - keep only sites for this TF from the GFF file
my $weight   = $opts{w}; # weight by which all TF energies are multiplied - does not affect Emean sites
my $offset   = $opts{offset}; # offset by which all TF energies can be shifted (incl. Emean)
####################################
my $out   = $opts{out}; # output landscape
$out = "default.en" if (!defined $out);
$Emean = 100.0 if (!defined $Emean);
####################################
$weight = 1.0 if (!defined $weight);
$offset = 0.0 if (!defined $offset);

# FASTA
my @seqs = read_fasta_noBio($fasta);
die "Please provide a single FASTA sequence!\n" unless (@seqs==1);
my $Lseq = length($seqs[0]);
my @dna_letters = split //, $seqs[0];
printf "Found %d bp in file: %s ..\n",$Lseq,$fasta;

# PWM
my @pwm = ();
my $DNAstep = read_pwm($pwm_file,\@pwm);

# Read in GFF file if given:
my @TF_start = (); # [start,end] of the TF sites from the GFF file
my @TF_end = ();
if (defined $gff) {
  my $TF_cnt = read_gff_single_tf($gff,\@TF_start,\@TF_end,$TFname,$chrN);
  printf "Found %d TF sites in %s [TFname,chrN] = [%s,%d] ..\n",$TF_cnt,$gff,$TFname,$chrN;
}

# Get max logscore for the PWM:
my $minE = get_minE(\@pwm);
#debug
#print "minE = $minE\n";

# Score all sites (or only those from the GFF file):
my @scores = ();
my @scores_rc = ();
PWMenergies(\@pwm,\@dna_letters,\@TF_start,\@TF_end,\@scores,\@scores_rc,$minE,$Emean);

# Generate the landscape:
open(V,">$out") or die "Can't open file: $!\n";
printf V "%6s  %7s  %7s\n","seqN","E","E_rc";
for my $j (0 .. ($#dna_letters - $#pwm) ) {
  printf V "%6d  %7.3f  %7.3f\n",$j+1,$scores[$j],$scores_rc[$j];
}
close V;

####################
###  Functions  ####
####################

sub getCommandLineOptions {
  use Getopt::Long;
  my $usage = qq{usage: $0 -f input.fna -pwm TF.pwm <-gff TFs.gff -chrN chr_num -TF TF_name -mean E.bkgr -w Eweight -offset Eoffset -out TF.etable>\n};
  # Get args
  my %opts = ();
  &GetOptions(\%opts, "f=s", "pwm=s", "gff=s", "chrN=i", "TF=s", "mean=f", "out=s", "w=f", "offset=f");
  if (!defined $opts{f} || !defined $opts{pwm}) {
    print STDERR "$usage\n";
    exit -1;
  }
  if (defined $opts{gff} && (!defined $opts{chrN} || !defined $opts{TF}) ) {
    print STDERR "If the GFF file is specified it must be accompanied by the TFname and the chrN!\n";
    print STDERR "$usage\n";
    exit -1;
  }
  &checkExist('f', $opts{f});
  &checkExist('f', $opts{pwm});
  &checkExist('f', $opts{gff}) if (defined $opts{gff});
  return %opts;
}

# checkExist()
#
sub checkExist {
  my ($type, $path) = @_;
  if ($type eq 'd') {
    if (! -d $path) {
      warn("dir not found: $path\n");
      exit -3;
    }
  }
  elsif ($type eq 'f') {
    if (! -f $path) {
      warn("file not found: $path\n");
      exit -3;
    }
    elsif (! -s $path) {
      warn("empty file: $path\n");
      exit -3;
    }
  }
}

#This sub loads fasta seqs into an array WITHOUT BioPerl
#functions. It can optionally perform various checks on
#the seqs it read.
sub read_fasta_noBio {
  my ($fasta_file,$mode) = @_;
  my @seqs = ();
  my $cnt = -1;
  open(II,"$fasta_file") or die "Can't open file: $!\n";
  while (<II>) {
    chomp;
    my @tt = split ' ';
    next if (@tt==0);
    if (substr($tt[0],0,1) eq '>') {
      $cnt++;
      next;
    }
    die "Unrecognized FASTA seq!\n" unless (@tt==1);
    if (!defined $seqs[$cnt]) {
      $seqs[$cnt] = uc($tt[0]); # make everything uppercase!!!
    }else{
      $seqs[$cnt] .= uc($tt[0]); # make everything uppercase!!!
    }
  }
  close II;
  if (defined $mode) {
    if ($mode eq "check_width") { # checks if all seqs are of the same length
      for my $i (1 .. $#seqs) {
	die "Seqs read from $fasta_file are not of the same length!\n"
	  if ((length($seqs[$i]) - length($seqs[$i-1]) ) != 0);
      }
    }
  }
  return @seqs;
}

#This aux. sub reverses and complements an input seq.
sub reverse_seq {
  my ($t) = @_;
  my %rev_bp = (
		"A" => "T",
		"C" => "G",
		"G" => "C",
		"T" => "A"
	       );
  my @aa = split //,$t;
  my @rev_aa = ();
  foreach my $a (reverse @aa) {
    push @rev_aa, $rev_bp{$a};
  }
  my $newt = join '', @rev_aa;
  return $newt;
}

# This sub reads TFs from a GFF tab-separated file (made by MacIsaac et al.).
# WARNING: make sure that the coordinates match up between genome versions!
# Also, make sure that the site is listed [begin,end] (end > begin) regardless
# of its orientation.
sub read_gff_single_tf {
  my ($file,$TF_start,$TF_end,$TFname,$chrN) = @_;
  # Check refs:
  if (ref($TF_start) ne 'ARRAY' || ref($TF_end) ne 'ARRAY') {
    die "Expected array/hash ref(s) in read_gff_single_tf!\nError in function args!\n";
  }
  # Read in file:
  open(I,"$file") or die "Can't open file: $!\n";
  my @lines = <I>;
  close I;
  # Parse lines:
  my $cnt = 0;
  foreach my $line (@lines) {
    chomp $line;
    my @tt = split /\t/,$line;
    next unless (@tt==9);
    #my $chrN_cur = $roman2arabic{$tt[0]}; # this is not necessary with TF GFF files
    my $chrN_cur = $tt[0];
    die "Illegal chrN: $tt[0]!\n" unless (defined $chrN_cur);
    next unless ($chrN_cur == $chrN);
    my $curname = $tt[8];
    $curname =~ s/Site //;
    $curname =~ s/\;//g;
    #die "Invalid TFname: $curname!\n" unless (length($curname)==4);
    next unless ($curname eq $TFname);
    push @$TF_start, $tt[3]; # 0-based!
    push @$TF_end, $tt[4]; # 0-based!
    $cnt++;
    #debug
    #printf "%s %d %d %d %1s\n",$tt[3],$tt[8],$tt[9],$tt[10],$tt[11];
    #exit;
  }
  return $cnt;
}

# Compute & return pairwise PWM log_scores:
sub PWMenergies {
  my ($pwm,$dna_letters,$start,$end,$scores,$scores_rc,$Emin,$Ebkgr) = @_;
  # Check refs:
  if (ref($dna_letters) ne 'ARRAY' || ref($pwm) ne 'ARRAY' ||
      ref($start) ne 'ARRAY' || ref($end) ne 'ARRAY' ||
      ref($scores) ne 'ARRAY' || ref($scores_rc) ne 'ARRAY') {
    die "Expected array/hash ref(s) in PWMenergies!\nError in function args!\n";
  }
  if ($#$start != $#$end) {
    die "Unequal [start,end] arrays in PWMenergies!\n";
  }
  my $et_length = $#$pwm + 1;
  if (!defined $gff) { # no GFF info was provided - [start,end] arrays must be empty!
    if ($#$start == -1 && $#$end == -1) { # empty [start,end] arrays - OK
      for (my $j=0; $j<=($#$dna_letters+1-$et_length); $j++) {
	push @$start, $j;
	push @$end, $j+$et_length-1;
      }
    }else{
      die "No GFF file but the [start,end] arrays are not empty!\n";
    }
  }
  for (my $j=0; $j<=($#$dna_letters+1-$et_length); $j++) {
    push @$scores, $Ebkgr + $offset;
    push @$scores_rc, $Ebkgr + $offset;
  }
  for my $j (0 .. $#$start) {
    die "start > end is illegal!\n" if ($start->[$j] > $end->[$j]); # start == end is OK, Lpwm = 1
    my @cursite = @$dna_letters[$start->[$j] .. $end->[$j]];
    #debug
    #print join("",@cursite),"\n";
    ##########
    my $curscore = score_site($pwm,\@cursite) - $Emin;
    #my $curscore = score_site($pwm,\@cursite);
    $scores->[$start->[$j]] = $weight * $curscore + $offset;
    ##########
    my $curscore_rc = score_site_rc($pwm,\@cursite) - $Emin;
    #my $curscore_rc = score_site_rc($pwm,\@cursite);
    $scores_rc->[$start->[$j]] = $weight * $curscore_rc + $offset;
    ##########
    #debug
    #printf "sc = %f; sc_rc = %f\n",$curscore,$curscore_rc;
  }
}

# This sub scores a single site.
sub score_site {
  my ($pwm,$site_bits) = @_;
  # Check refs:
  if (ref($pwm) ne 'ARRAY' || ref($site_bits) ne 'ARRAY') {
    die "Expected array ref(s) in score_site!\nError in function args!\n";
  }
  die "Length mismatch between PWM and site!\n" unless ($#$pwm == $#$site_bits);
  my $score = 0.0;
  for my $i (0 .. $#$pwm) {
    #$score += log($pwm->[$i][$base_map{$site_bits->[$i]}]);
    $score += log($pwm->[$i][$base_map{$site_bits->[$i]}]/$Pb[$base_map{$site_bits->[$i]}]);
  }
  return -$score; # flip to E!!
}

# This sub scores a revcom single site.
sub score_site_rc {
  my ($pwm,$site_bits) = @_;
  # Check refs:
  if (ref($pwm) ne 'ARRAY' || ref($site_bits) ne 'ARRAY') {
    die "Expected array ref(s) in score_site_rc!\nError in function args!\n";
  }
  die "Length mismatch between PWM and site!\n" unless ($#$pwm == $#$site_bits);
  my $score = 0.0;
  my $cnt_i =  $#$pwm;
  for my $i (0 .. $#$pwm) {
    #$score += log($pwm->[$cnt_i][$base_map{$base_rc{$site_bits->[$i]}}]);
    $score += log($pwm->[$cnt_i][$base_map{$base_rc{$site_bits->[$i]}}]/$Pb[$base_map{$base_rc{$site_bits->[$i]}}]);
    $cnt_i--;
  }
  return -$score; # flip to E!!
}

# This sub reads a single PWM in Saurabh's format.
sub read_pwm {
  my ($file,$pwm_ref) = @_;
  # Check refs:
  if (ref($pwm_ref) ne 'ARRAY') {
    die "Expected array ref(s) in read_pwm!\nError in function args!\n";
  }
  my $cnt = 0;
  my $flag = 0;
  open(I,"$file") or die "Can't open file: $!\n";
  while (<I>) {
    chomp;
    my @tt = split ' ';
    next if (@tt==0);
    if (substr($tt[0],0,1) eq '>') { # begin PWM
      $flag = 1;
      next;
    }elsif (substr($tt[0],0,1) eq '<') { # end PWM
      $flag = 0;
      last;
    }else{
      die "Unknown format of the entry line: expected at least 4 tokens!\n" if (@tt < 4);
      if ($flag==1) {
	for my $i (0 .. 3) {
	  $pwm_ref->[$cnt][$i] = $tt[$i] + $PSEUDO;
	}
	$cnt++;
      }
    }
  }
  close I;
  # Normalize:
  for my $i (0 .. $#$pwm_ref) {
    my $sum = 0.0;
    for my $j (0 .. 3) {
      $sum += $pwm_ref->[$i][$j];
    }
    for my $j (0 .. 3) {
      $pwm_ref->[$i][$j] /= $sum;
    }
  }
  return $cnt;
}

# This sub gets a minimum possible energy for the PWM
# (corresponding to the consensus seq).
sub get_minE {
  my ($pwm) = @_;
  #Check refs:
  if (ref($pwm) ne 'ARRAY') {
    die "Expected array ref(s) in get_max_scores!\nError in function args!\n";
  }
  my $maxscore = 0.0;
  for my $i (0 .. $#$pwm) {
    my @row_scores = ();
    for my $j (0 .. 3) {
      push @row_scores, log($pwm->[$i][$j]/$Pb[$j]);
    }
    my $max_row_score = max(\@row_scores);
    $maxscore += $max_row_score;
  }
  return -$maxscore; # note the "-" which makes it minE
}

# This function simply finds a maximum among all values in an input array,
# and returns it along with the array index.
sub max {
  my ($aref) = @_;
  #Check refs:
  if (ref($aref) ne 'ARRAY') {
    die "Expected array ref(s) in max!\nError in function args!\n";
  }
  my $max = -10000000000;
  my $maxind = undef;
  for my $j (0 .. $#$aref) {
    if ($aref->[$j] > $max) {
      $max = $aref->[$j];
      $maxind = $j;
    }
  }
  die "Could not find array max!\n" unless (defined $maxind);
  #return ($max,$maxind);
  return $max;
}

