#!/usr/bin/perl -w

use strict;

# This script is used to reset mean and variance of the input energy landscape.

# argv
my %opts = &getCommandLineOptions();
my $file = $opts{f}; # file with input energies for each position on the seq (both strands). Missing values are NOT allowed.
my $mean = $opts{mean}; # "standard" mean & var used in rescaling energies
my $var  = $opts{var}; # NOTE: both strands are rescaled with these
my $Bw   = $opts{bw}; # strand bias
my $scale = $opts{scale}; # scaling factor to mult all energies by (optional, happens AFTER mean/var rescaling)
my $Eout = $opts{Eout}; # output final (rescaled etc.) energies used in occupancy calculations
$Eout = $file.".rescaled" if (!defined $Eout);
$Bw  = 0.5 if (!defined $Bw);
$scale = 1.0 if (!defined $scale);

my @E = (); # energy landscape
my @seqN = (); # seq_num
my @meanEs = (); # corresponding <E> and var(E) (here just one number for each)
my @varEs = ();
read_energy_file($file,\@E,\@seqN,\@meanEs,\@varEs); # combines strands!

# Set mean/var if not defined:
$mean = $meanEs[0] if (!defined $mean);
$var = $varEs[0] if (!defined $var);

rescale_histogram(\@E,$mean,$var)
  unless (abs($mean - $meanEs[0]) < 1e-05 && abs($var - $varEs[0]) < 1e-05);

rescale_histogram_factor(\@E,$scale)
  unless (abs($scale - 1.0) < 1e-05);

# Recompute mean & var before outputting rescaled energies:
$mean = ave(\@E);
$var = var(\@E);

# Output energies:
open(V,">$Eout") or die "Can't open file: $!\n";
printf V "# Bw = %6.2f\n",$Bw;
printf V "# <E> = %6.2f, var(E) = %6.2f\n",$mean,$var;
print V "    seqN       E\n";
for my $k (0 .. $#E) {
  printf V "%8d   %6.3f\n",$seqN[$k],$E[$k];
}
close V;

####################
###  Functions  ####
####################

sub getCommandLineOptions {
  use Getopt::Long;
  my $usage = qq{usage: $0 -f energies.dat <[-mean X AND/OR -var Y] -Eout energies.dat.rescaled -scale Z -bw strand_bias>\n};
  # Get args
  my %opts = ();
  &GetOptions(\%opts, "f=s", "Eout=s", "mean=f", "var=f", "bw=f", "scale=f");
  if (!defined $opts{f}) {
    print STDERR "$usage\n";
    exit -1;
  }
  #if (!defined $opts{mean} && !defined $opts{var}) {
  #  print STDERR "ERROR: at least one of the -mean & -var options must be specified!\n";
  #  exit -1;
  #}
  &checkExist('f', $opts{f});
  return %opts;
}

# checkExist()
#
sub checkExist {
  my ($type, $path) = @_;
  if ($type eq 'd') {
    if (! -d $path) {
      warn("dir not found: $path\n");
      exit -3;
    }
  }
  elsif ($type eq 'f') {
    if (! -f $path) {
      warn("file not found: $path\n");
      exit -3;
    }
    elsif (! -s $path) {
      warn("empty file: $path\n");
      exit -3;
    }
  }
}

# This function simply finds an average value for all values in an input array.
sub ave {
  my ($aref) = @_;
  #Check refs:
  if (ref($aref) ne 'ARRAY') {
    die "Expected array ref(s) in ave!\nError in function args!\n";
  }
  my $ave = 0.0;
  if (scalar(@$aref) > 0) {
    for my $j (0 .. $#$aref) {
      $ave += $aref->[$j];
    }
    $ave /= ($#$aref+1);
  }
  return $ave;
}

sub covar {
  my ($l1_ref,$l2_ref) = @_;
  # Check refs:
  if (ref($l1_ref) ne 'ARRAY' || ref($l2_ref) ne 'ARRAY') {
    die "Expected array ref in covar!\n";
  }
  die "Arrays of unequal length!\n" if ($#$l1_ref != $#$l2_ref);
  my $covar = 0.0;
  for my $i (0 .. $#$l1_ref) {
    $covar += ($l1_ref->[$i] * $l2_ref->[$i]);
  }
  $covar /= scalar(@$l1_ref);
  my $m1 = ave($l1_ref);
  my $m2 = ave($l2_ref);
  
  return ($covar - $m1*$m2);
}

sub var {
  my ($list_ref) = @_;
  # Check refs:
  if (ref($list_ref) ne 'ARRAY') {
    die "Expected array ref in var!\n";
  }
  my $var = covar($list_ref,$list_ref);
  
  return $var;
}

# This sub reads in a file with rescaled energies [2 columns: seqN  E].
# Note that it does not fill in any missing values!!
sub read_energy_file {
  my ($file,$E,$seqN,$meanE,$varE) = @_;
  # Check refs:
  if (ref($E) ne 'ARRAY' || ref($seqN) ne 'ARRAY' ||
      ref($meanE) ne 'ARRAY' || ref($varE) ne 'ARRAY') {
    die "Expected array ref(s) in read_energy_file!\nError in function args!\n";
  }
  print "Reading file: $file ..\n";
  # Read in file:
  open(I,"$file") or die "Can't open file: $!\n";
  my @lines = <I>;
  close I;
  # Skip comments header (arbitrary number of lines marked with '#'):
  my $header_cnt = 0;
  while (1) {
    chomp $lines[$header_cnt];
    if (substr($lines[$header_cnt],0,1) ne '#') {
      last;
    }else{
      $header_cnt++;
    }
  }
  # Find tokens in the header line:
  chomp $lines[$header_cnt];
  my @toks = split ' ',$lines[$header_cnt];
  my ($ind_seqN,$ind_E,$ind_E_rc);
  for my $i (0 .. $#toks) { # note that we do not check for multiple instances in label names!!
    $ind_seqN = $i if ($toks[$i] eq "seqN");
    $ind_E = $i if ($toks[$i] eq "E" || $toks[$i] eq "Ewo"); # we do not expect to find both E & Ewo in one file ..
    $ind_E_rc = $i if ($toks[$i] eq "E_rc" || $toks[$i] eq "Ewo_rc");
  }
  unless (defined $ind_seqN && defined $ind_E) {
    die "Could not find all required columns in file: $file!\n";
  }
  if (!defined $ind_E_rc) {
    print STDERR "Looks like landscape read from file: $file does not have revcom energies!\nStrand energies will not be combined ..\n";
  }
  # Store energies:
  my $cnt = 0;
  my @Eloc = ();
  my @Eloc_rc = ();
  for my $i ($header_cnt+1 .. $#lines) { # skip header
    chomp $lines[$i];
    my @tt = split ' ',$lines[$i];
    next if (@tt==0);
    #$E->{$tt[$ind_seqN]} = $tt[$ind_E];
    push @$seqN, $tt[$ind_seqN];
    push @Eloc, $tt[$ind_E];
    push @Eloc_rc, $tt[$ind_E_rc] if (defined $ind_E_rc);
    $cnt++;
  }
  if (defined $ind_E_rc) {
    my @Ecomb = ();
    combine_strands(\@Eloc,\@Eloc_rc,\@Ecomb);
    foreach my $en (@Ecomb) {
      push @$E, $en;
    }
  }else{
    foreach my $en (@Eloc) {
      push @$E, $en;
    }
  }
  printf "Found %d values in file: %s ..\n",$cnt,$file;
  push @$meanE, ave($E);
  push @$varE, var($E);
  printf "Moments of E distribution: (<E>,var(E)) = (%.3f,%.3f) ..\n",$meanE->[$#$meanE],$varE->[$#$varE];
}

# This sub converts energies for 2 strands: (E1,E2) into a single statistical weight:
# Bw*w1 + \bar{Bw}*w2, where wi = exp(-Ei).
# The idea is that this is the appropriate way of combining the energies from both strands
# *before* Z is computed (see accompanying theory).
# Physically, changing (Bw + \bar{Bw}) amounts to shifting the zero of all energies and should not
# affect any probabilities, so we simply choose Bw + \bar{Bw} = 1.
# Note that we report -log(Bw*w1 + (1-Bw)*w2) because it is the energies that are expected
# by the rest of the program.
sub combine_strands {
  my ($E1,$E2,$E) = @_;
  # Check refs:
  if (ref($E1) ne 'ARRAY' || ref($E2) ne 'ARRAY' || ref($E) ne 'ARRAY') {
    die "Expected hash ref(s) in combine_strands!\nError in function args!\n";
  }
  print "Combining strand energies (Bw = $Bw) ..\n";
  # Keep mean and var of the original distribution on strand 1 (this is an arbitrary choice):
  #my $mean = $Bw*ave($E1) + (1-$Bw)*ave($E2);
  #my $var = $Bw*var($E1) + (1-$Bw)*var($E2);
  # Compute (log) probabilities by combining energies from 2 strands:
  for my $k (0 .. $#$E1) { # common offset & scaling factor
    my $p1 = exp(-$E1->[$k]);
    my $p2 = exp(-$E2->[$k]);
    $E->[$k] = -log($Bw*$p1 + (1-$Bw)*$p2);
    # debug
    #printf "%d    %.3f\n",$k,$E->[$k];
  }
  # Restore original mean & var after combining strands - switch off (AM 12/06) -> unwanted side effects with TF landscapes!
  #rescale_histogram($E,$mean,$var);
  #debug
  #exit;
}

# This function rescales a distribution (ie shifts its mean and rescales its variance)
# using input mean & var (either user-provided or inferred from another, "standard" distribution).
# NOTE: this function should probably be used BEFORE combining energies from 2 strands,
# because rescaling affects relative contributions from 2 strands.
sub rescale_histogram {
  my ($Eref,$mean,$var) = @_;
  #Check refs:
  if (ref($Eref) ne 'ARRAY') {
    die "Expected array/hash ref(s) in rescale_histogram!\nError in function args!\n";
  }
  my $aveE = ave($Eref);
  my $varE = var($Eref);
  printf "Rescaling the E distribution: (<E>,var(E)) = (%.3f,%.3f) to (<E>,var(E)) = (%.3f,%.3f) ..\n",
    $aveE,$varE,$mean,$var;
  my $scaling_factor;
  if ($varE > 1e-05) {
    $scaling_factor = sqrt($var/$varE);
  }else{
    print STDERR "Warning: zero variance [varE] cannot be rescaled - setting scaling factor to 1.0!\n";
    $scaling_factor = 1.0;
  }
  foreach my $k (0 .. $#$Eref) { # common offset & scaling factor
    $Eref->[$k] -= $aveE;
    $Eref->[$k] *= $scaling_factor;
    $Eref->[$k] += $mean;
  }
  #debug
  #printf "New (<E>,var(E)) = (%.3f,%.3f) ..\n",ave($Eref),var($Eref);
  #exit;
}

# This function rescales a distribution with a scaling factor: E_i -> E_i * factor
sub rescale_histogram_factor {
  my ($Eref,$scaling_factor) = @_;
  #Check refs:
  if (ref($Eref) ne 'ARRAY') {
    die "Expected array/hash ref(s) in rescale_histogram_factor!\nError in function args!\n";
  }
  my $aveE = ave($Eref);
  my $varE = var($Eref);
  printf "Rescaling the E distribution: (<E>,var(E)) = (%.3f,%.3f) to (<E>,var(E)) = (%.3f,%.3f) ..\n",
    $aveE,$varE,$aveE,$scaling_factor*$scaling_factor*$varE;
  foreach my $k (0 .. $#$Eref) { # common offset & scaling factor
    $Eref->[$k] -= $aveE;
    $Eref->[$k] *= $scaling_factor;
    $Eref->[$k] += $aveE;
  }
  #debug
  #printf "New (<E>,var(E)) = (%.3f,%.3f) ..\n",ave($Eref),var($Eref);
  #exit;
}

