#!/usr/bin/perl -w

use strict;

# This aux. script is used to run dynapro over a set of directories.

my $HOME = $ENV{HOME};
my $DYNAPRO = $HOME."/DynaPro_1.1/bin"; # home dir for the C++ dynamic programming code

#my $Lembed = 5;

# argv
my %opts = &getCommandLineOptions();
my $dirlist     = $opts{l}; # input list of dirs (1 token per line, rest of line ignored)
my $occ_ext     = $opts{occ_ext}; # occupancy filename extension, eg ".occ.3" [now with dot in front!]
my $et_ext      = $opts{et_ext}; # etable filename extension, eg ".etable.rescaled.2" [now with dot in front!]
my $bkgr_choice = $opts{bkgr}; # uniform/genomic/none bkgr nuc freqs, used in embedding sites & creating the bkgr landscape
my $Pmin        = $opts{enforceP}; # enforce Pmin for each type of object
my $Lembed      = $opts{Lem};

#####
my @et_ext_toks = ();
if (defined $et_ext) {
  @et_ext_toks = split ' ',$et_ext;
}else{
  $et_ext_toks[0] = ".etable.rescaled.2";
}
#####
my @Lembed_toks = ();
if (defined $Lembed) {
  @Lembed_toks = split ' ',$Lembed;
}else{
  $Lembed_toks[0] = 5;
}
#####
my @Pmin_toks = ();
if (defined $Pmin) {
  @Pmin_toks = split ' ',$Pmin;
}
#####
#$Lembed = 5 if (!defined $Lembed);
$occ_ext = ".occ.3" if (!defined $occ_ext);
#$et_ext = ".etable.rescaled.2" if (!defined $et_ext);
$bkgr_choice = "uniform" if (!defined $bkgr_choice);

foreach my $L (@Lembed_toks) {
  die "Illegal embedding length!\n" if ($L < 0);
}

if (defined $Pmin) {
  die "Array mismatch: please provide 1 et_ext, 1 Lembed and 1 Pmin for each type of input object!\n"
    unless ($#et_ext_toks == $#Lembed_toks && $#et_ext_toks == $#Pmin_toks);
}else{
  die "Array mismatch: please provide 1 et_ext and 1 Lembed for each type of input object!\n"
    unless ($#et_ext_toks == $#Lembed_toks);
}

my @dirs = read_list($dirlist);

foreach my $dd (@dirs) {
  chomp $dd;
  chdir $dd or die "Can't cd: $!\n";
  print "***** Working in $dd .. *****\n";
  my $fasta = $dd.".fna"; # FASTA input
  &checkExist('f', $fasta);
  my $dynapro_file = "etables_tmp.dat";
  open(XA,">$dynapro_file") or die "Can't open file: $!\n";
  for my $i (0 .. $#et_ext_toks) {
    my $efile = $dd.$et_ext_toks[$i]; # energy landscape input
    &checkExist('f', $efile);
    #############
    printf XA "%s  %2d\n",$efile,$Lembed_toks[$i];
  }
  close XA;
  my $occfile = $dd.$occ_ext; # single output occ file
  #############
  if (defined $Pmin) {
    #############
    my $Pmin_file = "Pmin_tmp.dat";
    open(XB,">$Pmin_file") or die "Can't open file: $!\n";
    foreach my $pmin (@Pmin_toks) {
      printf XB "%f\n",$pmin;
    }
    close XB;
    #############
    system("$DYNAPRO/dynapro -fasta $fasta -et $dynapro_file -out $occfile -bkgr $bkgr_choice -enforceP $Pmin_file");
    `/bin/rm $Pmin_file` if (-e $Pmin_file);
  }else{
    system("$DYNAPRO/dynapro -fasta $fasta -et $dynapro_file -out $occfile -bkgr $bkgr_choice");
  }
  &checkExist('f', $occfile);
  `/bin/rm $dynapro_file` if (-e $dynapro_file);
  print "*****************************\n";
  chdir '..' or die "Can't cd: $!\n";
}

####################
###  Functions  ####
####################

sub getCommandLineOptions {
  use Getopt::Long;
  my $usage = qq{usage: $0 -l dir.list <-et_ext \"etable_extension1 etable_extension2 ..\" -occ_ext occ_extension -Lem \"Lembed1 Lembed2 ..\" -bkgr uniform/genomic/none -enforceP \"Pmin1 Pmin2 ..\">\n};
  # Get args 
  my %opts = ();
  &GetOptions (\%opts, "l=s", "et_ext=s", "occ_ext=s", "Lem=s", "bkgr=s", "enforceP=s");
  # Check for legal invocation
  if (!defined $opts{l}) {
    print STDERR "$usage\n";
    exit -1;
  }
  &checkExist('f', $opts{l});
  return %opts;
}

# checkExist()
#
sub checkExist {
  my ($type, $path) = @_;
  if ($type eq 'd') {
    if (! -d $path) {
      warn("dir not found: $path\n");
      exit -3;
    }
  }
  elsif ($type eq 'f') {
    if (! -f $path) {
      warn("file not found: $path\n");
      exit -3;
    }
    elsif (! -s $path) {
      warn("empty file: $path\n");
      exit -3;
    }
  }
}

# This function reads in a list of items from a file (one item per line, rest of line ignored;
# '#' comment symbols recognized).
sub read_list {
  my ($file) = @_;
  my @tokens = ();
  my $cnt = 0;
  open(I,"$file") or die "Can't open file: $!\n";
  while (<I>) {
    chomp;
    my @tt = split ' ';
    next if (@tt==0);
    next if (substr($tt[0],0,1) eq '#');
    my $token = $tt[0];
    push @tokens,$token;
    $cnt++;
  }
  close I;
  print "Read $cnt token(s) from file: $file ..\n";
  return @tokens;
}

