/* dynapro.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DYNAPRO implements a dynamic programming algorithm for predicting
DNA occupancies by arbitrary objects whose interaction energies with DNA are
known for each possible binding position.

The DYNAPRO method and its applications are described in the following
paper:

******************************************************************/

#include "options.h"
#include "fastafile.h"
#include "sequence.h"
#include "landscapes.h"
#include "stats.h"

// This main function implements a dynamic programming algorithm which translates
// energy landscapes for an arbitrary number of DNA-binding objects (such as histone
// octamers) into starting probabilities and occupancies.

int main(int argc, char **argv) {
  
  // (A) ==== Read in command-line options: ==== //
  if (argc < 2) {
    fprintf(stderr,"usage: %s -fasta <Fasta file with input seq> -et <landscapes.dat> [-out <table.occ>] [-bw <strand_bias>] [-bkgr <uniform/genomic/none>] [-enforceP <Pmin.dat>]\n",argv[0]);
    exit(1);
  }
  int argbase = 1;
  struct Options *opt = ReadArguments(argbase, argc, argv);
  
  // (B) ==== Read in the sequence(s) from input file: ==== //
  char *fastafile = opt->fasta;
  printf("=================================\n");
  printf("FASTA file: %s\n",fastafile);
  printf("Strand bias: %.2f\n",opt->bw);
  printf("Background: %s\n",opt->bkgr.c_str() );
  FastaFile sequences;
  sequences.ReadFasta(fastafile);
  std::cerr << "Read "<<sequences.Size()<<" seq(s) from file: "<<fastafile<<std::endl;
  if (sequences.Size() != 1) {
    std::cerr << "ERROR: fasta file "<<fastafile<<" must contain a single sequence!\nExiting ..\n";
    exit(EXIT_SUCCESS);
  }
  Sequence *seq = sequences[0];
  printf("Found %d bp in file: %s ..\n",seq->Length(),fastafile);
  
  // (C) ==== Process energy landscapes and read in Lembed (if provided): ==== //
  Bkgr bkgr(opt->bkgr); // bkgr prms
  Landscapes en(opt->et,opt->bw,bkgr,seq); // energy landscapes, incl. bkgr
  
  printf("Output file with occupancies and probabilities: %s\n",opt->out.c_str() );
  printf("=================================\n");
  
  // (D) Read in Pmin values (if any):
  Probmin Pmin(opt->enforceP,en.GetNLAND()-1);
  
  // (E) Do the dynamic programming:
  en.dynapro(seq->Length(),Pmin);
  //debug
  //exit(0);
  
  // (F) Output the result:
  en.output_prob_occ(opt->out,opt->bkgr,seq->Length(),opt->bw);
  
  delete opt;
}

