/* landscapes.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DYNAPRO implements a dynamic programming algorithm for predicting
DNA occupancies by arbitrary objects whose interaction energies with DNA are
known for each possible binding position.

The DYNAPRO method and its applications are described in the following
paper:

******************************************************************/

#include "landscapes.h"

// A generic function to read 1 item per line from a file and return a vector with read items.
// No header is expected.
std::vector<std::string> ReadList(std::string const &filename) {
  std::vector<std::string> items;
  
  using std::cout;
  using std::cerr;
  using std::endl;
  using std::ifstream;
  
  ifstream file(filename.c_str() );
  
  char buffer[512];
  char citem[512];
  
  if (file.fail()) {
    cerr << "Could not open file: "<<filename<<" for reading!\nExiting..\n";
    exit(EXIT_FAILURE);
  }
  
  int cnt = 0; // item counter
  while (!file.eof()) {
    file.getline(buffer,512,'\n');
    if (file.fail() && !file.eof() ) file.clear();
    if (buffer[0] == '#') continue; // skip commented lines
    if (sscanf(buffer,"%s",citem) == 1) {
      std::string item(citem);
      items.push_back(item);
      cnt++;
    }
  }
  file.close();
  cerr << "Read "<<cnt<< " item(s) from file: "<<filename<<" ..\n";
  
  return items;
}

// A generic function to read 2 items per line from a file and return two vectors with read items.
// No header is expected.
void ReadList_double(std::string const &filename, std::vector<std::string> &c1, std::vector<int> &c2) {
  
  using std::cout;
  using std::cerr;
  using std::endl;
  using std::ifstream;
  
  ifstream file(filename.c_str() );
  
  char buffer[512];
  char item1[256];
  int item2;
  
  if (file.fail()) {
    cerr << "Could not open file: "<<filename<<" for reading!\nExiting..\n";
    exit(EXIT_FAILURE);
  }
  
  int cnt = 0; // item counter
  while (!file.eof()) {
    file.getline(buffer,512,'\n');
    if (file.fail() && !file.eof() ) file.clear();
    if (buffer[0] == '#') continue; // skip commented lines
    if (sscanf(buffer,"%s %d",item1,&item2) == 2) {
      std::string item1_str(item1);
      c1.push_back(item1_str);
      c2.push_back(item2);
      cnt++;
    }
  }
  file.close();
  cerr << "Read "<<cnt<< " item pair(s) from file: "<<filename<<" ..\n";
}

////////////////////////

Bkgr::Bkgr() {
  for (unsigned int i=0; i<NUC_NUM; ++i) {
    Eb[i] = 0.0;
  }
}

Bkgr::Bkgr(std::string const &bkgr_choice) {
  if (bkgr_choice == "uniform") {
    for (unsigned int i=0; i<NUC_NUM; ++i)
      Eb[i] = -log(0.25);
  }else if (bkgr_choice == "genomic") {
    Eb[0] = -log(0.309012);
    Eb[1] = -log(0.191674);
    Eb[2] = -log(0.191303);
    Eb[3] = -log(0.308012);
  }else if (bkgr_choice == "none") {
    for (unsigned int i=0; i<NUC_NUM; ++i)
      Eb[i] = 0.0;
  }else{
    fprintf(stderr,"Unknown background option: %s!\n",bkgr_choice.c_str());
    exit(EXIT_FAILURE);
  }
}

double Bkgr::GetEb(unsigned int ind) {
  if (ind >= 0 && ind <= 3) {
    return Eb[ind];
  }else{
    fprintf(stderr,"Background index: %d is out of range!\n",ind);
    exit(EXIT_FAILURE);
  }
}

double Bkgr::GetEb(char nuc) {
  if (nuc == 'A' || nuc == 'a') {
    return Eb[0];
  }else if (nuc == 'C' || nuc == 'c') {
    return Eb[1];
  }else if (nuc == 'G' || nuc == 'g') {
    return Eb[2];
  }else if (nuc == 'T' || nuc == 't') {
    return Eb[3];
  }else{
    fprintf(stderr,"Unknown nucleotide: %c!\n",nuc);
    exit(EXIT_FAILURE);
  }
}

////////////////////////

Probmin::Probmin() {
  pmin = NULL;
}

Probmin::~Probmin() {
  delete [] pmin;
}

// NOTE: L is the number of *non-bkgr* (ie input) landscapes.
Probmin::Probmin(std::string const &Pmin_choice, unsigned int L) {
  Lpmin = L;
  pmin = new double[L];
  if (Pmin_choice == "none") {
    for (unsigned int i=0; i<L; ++i) {
      pmin[i] = -1.0;
    }
  }else{
    std::vector<std::string> tokens = ReadList(Pmin_choice);
    if (L != tokens.size() ) {
      std::cerr << "Expected "<<L<<" Pmin values but found "<<tokens.size()<<"!\nExiting ..\n";
      exit(EXIT_FAILURE);
    }
    for (unsigned int i=0; i<L; ++i) {
      pmin[i] = atof(tokens.at(i).c_str() );
    }
  }
}

double Probmin::GetPmin(unsigned int ind) {
  if (ind >= 0 && ind < Lpmin) {
    return pmin[ind];
  }else{
    std::cerr << "Index "<<ind<<" is out of range!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
}

unsigned int Probmin::GetLPmin(void) {
  return Lpmin;
}

///////////////////////

Landscapes::Landscapes() {
  NLAND = 0;
}

Landscapes::~Landscapes() {}

Landscapes::Landscapes(std::string const &filename, double Bw, Bkgr &bkgr, Sequence *seq) {
  
  if (filename == "none") {
    std::cerr << "Error: no file with etable names provided!\n";
    exit(EXIT_FAILURE);
  }
  
  std::vector<std::string> landscape_names;
  Lembed.push_back(0); // no embedding for bkgr
  ReadList_double(filename,landscape_names,Lembed); // read filenames with landscapes and Lembed offsets
  
  printf("Found %d energy landscape file(s):\n",landscape_names.size() );
  if (landscape_names.size() == 0) {
    fprintf(stderr,"ERROR: please provide at least one energy landscape!\nExiting..\n");
    exit(EXIT_FAILURE);
  }
  
  NLAND = landscape_names.size() + 1; // +1 for the bkgr landscape ..
  
  en.resize(NLAND);
  en_rc.resize(NLAND);
  seqN.resize(NLAND);
  
  prob.resize(NLAND-1);
  occ.resize(NLAND-1);
  
  // Bkgr landscape (flat for now but could be modified to any shape fixed on the seq):
  build_bkgr(bkgr,seq);
  
  // Energy landscapes from input files (nucleosome + TFs):
  for (unsigned int i=0; i<landscape_names.size(); ++i) {
    printf("%s\n",landscape_names[i].c_str() );
    read_energy_file_processed(landscape_names[i],i+1);
  }
  
  // Combine strands if both E and Erc were read in (except bkgr):
  for (unsigned int i=1; i<NLAND; ++i) { // skip bkgr
    //debug
    //printf("size=%d\n",en_rc[i].size() );
    if (en_rc[i].size() > 0) { // E_rc exists
      combine_strands(en[i], en_rc[i], Bw);
    }
  }
  
  // Compute <E> and var(E) for all landscapes (before embedding!):
  for (unsigned int i=0; i<NLAND; ++i) { // over all landscapes ..
    double aveE = ave(en[i]);
    double varE = var(en[i]);
    meanEs.push_back(aveE);
    varEs.push_back(varE);
    printf("Moments of E distribution: (<E>,var(E)) = (%.3f,%.3f) ..\n",aveE,varE);
  }
  
  // Embed landscapes in background (if requested):
  for (unsigned int j=1; j<NLAND; ++j) { // over all landscapes except bkgr ..
    embed_landscape(bkgr,seq,j);
  }
  
  // Determine footprints (in bp) corresponding to the input energy landscapes:
  get_footprints((unsigned int) seq->Length() );
  
}

unsigned int Landscapes::GetNLAND(void) {
  return NLAND;
}

// This function is used to build the bkgr landscape.
void Landscapes::build_bkgr(Bkgr &bkgr, Sequence *seq) {
  for (unsigned int i=0; i<seq->Length(); ++i) {
    char nuc = seq->CharAt(i);
    en[0].push_back(bkgr.GetEb(nuc) );
    seqN[0].push_back(i+1);
  }
}

// This sub reads in a file with rescaled energies [2 columns: seqN  E].
//  Note that it does not fill in any missing values!!
void Landscapes::read_energy_file_processed(std::string const &etname, unsigned int ind) {
  
  using std::cout;
  using std::cerr;
  using std::endl;
  using std::ifstream;
  
  ifstream file(etname.c_str() );
  
  char buffer[1024];
  //char item1[256],item2[256],item3[256];
  int item1;
  double item2,item3;
  
  if (file.fail()) {
    cerr << "Could not open file: "<<etname<<" for reading!\nExiting..\n";
    exit(EXIT_FAILURE);
  }
  
  int cnt = 0; // item counter
  while (!file.eof()) {
    file.getline(buffer,1024,'\n');
    if (file.fail() && !file.eof() ) file.clear();
    
    if (buffer[0] == '#') continue; // skip commented lines
    if (strstr(buffer,"seqN") ) continue; // skip header
    
    if (sscanf(buffer,"%d %lf",&item1,&item2) == 2) { // seqN  E
      if (cnt==0) {
	cout << "Looks like landscape "<<ind<<" (read from file: "<<etname<<") does not have revcom energies!\nStrand energies will not be combined ..\n";
      }
      seqN[ind].push_back(item1);
      en[ind].push_back(item2);
      //debug
      //printf("%d  %lf\n",item1,item2);
      cnt++;
    }else if (sscanf(buffer,"%d %lf %lf",&item1,&item2,&item3) == 3) { // seqN  E  E_rc
      seqN[ind].push_back(item1);
      en[ind].push_back(item2);
      en_rc[ind].push_back(item3);
      cnt++;
    }
  }
  file.close();
  printf("Found %d values in file: %s ..\n",cnt,etname.c_str() );
}

// This sub converts energies for 2 strands: (E1,E2) into a single statistical weight:
// Bw*w1 + \bar{Bw}*w2, where wi = exp(-Ei).
// The idea is that this is the appropriate way of combining the energies from both strands
// *before* Z is computed (see accompanying theory).
// Physically, changing (Bw + \bar{Bw}) amounts to shifting the zero of all energies and should not
// affect any probabilities, so we simply choose Bw + \bar{Bw} = 1.
// Note that we report -log(Bw*w1 + (1-Bw)*w2) because it is the energies that are expected
// by the rest of the program.
// NOTE: the function will erase data in Erc and reload E (!).
void Landscapes::combine_strands(std::vector<double> &E, std::vector<double> &Erc, double Bw) {
  // Check if the E and Erc arrays are of equal size:
  if (E.size() != Erc.size() ) {
    std::cerr << "Arrays of unequal size cannot be combined!\n";
    exit(EXIT_FAILURE);
  }
  // Keep mean and var of the original distribution on strand 1 (this is an arbitrary choice):
  double mean2 = 0.5 * (ave(E) + ave(Erc) );
  double var2 = 0.5 * (var(E) + var(Erc) );
  // Compute (log) probabilities by combining energies from 2 strands:
  for (unsigned int k=0; k<E.size(); ++k) {
    double p1 = exp(-E[k]);
    double p2 = exp(-Erc[k]);
    E[k] = -log(Bw*p1 + (1-Bw)*p2); // reloads E!
  }
  // Restore original mean & var after combining strands:
  rescale_histogram(E,mean2,var2);
  // Remove all elements from the Erc array:
  Erc.clear();
}

// This function rescales a distribution (ie shifts its mean and rescales its variance)
// using input mean & var (either user-provided or inferred from another, "standard" distribution).
// NOTE: this function should probably be used BEFORE combining energies from 2 strands,
// because rescaling affects relative contributions from 2 strands.
void Landscapes::rescale_histogram(std::vector<double> &E, double mean2, double var2) {
  double aveE = ave(E);
  double varE = var(E);
  printf("Rescaling the E distribution: (<E>,var(E)) = (%.3f,%.3f) to (<E>,var(E)) = (%.3f,%.3f) ..\n",
	 aveE,varE,mean2,var2);
  double scaling_factor = sqrt(varE/var2);
  for (unsigned int k=0; k<E.size(); ++k) {
    E[k] -= aveE;
    E[k] /= scaling_factor;
    E[k] += mean2;
  }
}

// This sub determines the footprints (in bp) for all input objects.
void Landscapes::get_footprints(unsigned int Lseq) {
  using std::cout;
  using std::cerr;
  using std::endl;
  
  if (Lseq <= 0) {
    cerr << "Illegal Lseq: "<<Lseq<<"!\n";
    exit(EXIT_FAILURE);
  }
  for (unsigned int i=0; i<NLAND; ++i) {
    if (seqN[i].at(0) != 1) { // all seqnums should be one-based for consistency!!
      cerr << "First position on energy landscape "<<i<<" does not have seqN = 1!\n";
      exit(EXIT_FAILURE);
    }
    // Check for gaps:
    for (unsigned int j=1; j<seqN[i].size(); ++j) {
      if (seqN[i][j] - 1 != seqN[i][j-1]) {
	cerr << "Energy landscape "<<i<<" has a gap in seqN: ["<<j<<","<<j-1<<"]!\n";
	exit(EXIT_FAILURE);
      }
    }
    // Determine footprint:
    int Lcur = Lseq - seqN[i][seqN[i].size()-1] + 1;
    if (Lcur < 0) {
      cerr << "Something is wrong: the footprint of the object with energy landscape "<<i<<" is "<<Lcur<<" <= 0!\n";
      exit(EXIT_FAILURE);
    }
    printf("Looks like the footprint of the object with energy landscape %d equals %d bp ..\n",i,Lcur);
    Ls.push_back(Lcur);
    //debug
    //printf("L=%d\n",Lcur);
  }
}

// This sub does the dynamic programming, ie computes the partition function, posterior
// probabilities and occupancies for all input objects.
void Landscapes::dynapro(int Lseq, Probmin &Pmin) {
  using std::cout;
  using std::cerr;
  using std::endl;
  
  if (Ls.size() != Lembed.size() ) {
    cerr << "Mismatch in the number of objects and their embedding extensions!\n";
    exit(EXIT_FAILURE);
  }
  
  // Forward:
  std::vector<double> Forw;
  for (int i=0; i<=Lseq; ++i) {
    Forw.push_back(0.0);
  }
  for (int i=1; i<=Lseq; ++i) { // over all one-based bp numbers ..
    double Factor = 0.0;
    for (int j=0; j<NLAND; ++j) { // over all landscapes ..
      if ((i - Ls[j] + 1) > 0) {
	double Ej = en[j][i - Ls[j]];
	//debug
	//printf("%d  Lj=%d\n",j,(i - Ls[j] + 1));
	//printf("%d  Ej=%f\n",i,Ej);
	Factor += exp(Forw[i - Ls[j]] - Forw[i-1] - Ej);
      }
    }
    Forw[i] = Forw[i-1] + log(Factor);
    //debug
    //printf("%d  %f\n",i,Forw[i]);
  }
  //debug
  //exit(0);
  
  // Backward:
  std::vector<double> Back;
  for (int i=1; i<=(Lseq+1); ++i) {
    Back.push_back(0.0);
  }
  for (int i=Lseq; i>=1; --i) {
    double Factor = 0.0;
    for (int j=0; j<NLAND; ++j) { // over all landscapes ..
      if ((Lseq - Ls[j] + 2 - i) > 0) {
	double Ej = en[j][i - 1];
	Factor += exp(Back[i + Ls[j]] - Back[i+1] - Ej);
      }
    }
    Back[i] = Back[i+1] + log(Factor);
  }
  
  // Partition function:
  printf("# logF = %.4f, logR = %.4f\n",Forw[Lseq],Back[1]);
  
  // Posterior probability:
  for (int j=1; j<NLAND; ++j) { // over all landscapes except bkgr ..
    for (int i=1; i<=(Lseq - Ls[j] + 1); ++i) {
      prob[j-1].push_back(exp(Forw[i-1] - en[j][i-1] + Back[i + Ls[j]] - Back[1]) );
    }
    for (int i=(Lseq - Ls[j] + 2); i<=Lseq; ++i) { // pad with 0's on the right
      prob[j-1].push_back(0.0);
    }
  }
  
  // Remove all probabilites below Pmin thresholds:
  for (int i=0; i<NLAND-1; ++i) {
    for (int j=0; j<prob[i].size(); ++j) {
      if (prob[i][j] < Pmin.GetPmin(i) ) prob[i][j] = 0.0;
    }
  }
  
  // Occupancy by object j (the probability that position i is covered by *any* object of type j):
  // NOTE: occupancy is computed accounting for the fact that the actual object is embedded in the
  // center of a longer composite object: Lem+L+Lem (where Lem=0 is allowed).
  for (int j=1; j<NLAND; ++j) { // over all landscapes except bkgr ..
    for (int i=1; i<=Lseq; ++i) { // over all basepairs ..
      // A cheaper way to compute occupancies:
      double occ_cur_bp = 0.0;
      if ((i - 1 - Lembed[j]) >= 0)
	occ_cur_bp += prob[j-1][i - 1 - Lembed[j]]; // prob at basepair i
      if ((i-2) >= 0)
	occ_cur_bp += occ[j-1][i-2]; // + previous occ (at i-1) [ if exists ]
      if ((i - Ls[j] + Lembed[j] - 1) >= 0)
	occ_cur_bp -= prob[j-1][i - Ls[j] + Lembed[j] - 1]; // - prob of the object which no longer extends to i [ if exists ]
      occ[j-1].push_back(occ_cur_bp);
    }
  }
  
}

// This function embeds objects in bkgr seq, e.g. 5+147+5 for the nucleosome.
// The idea is that the nucleosomes are separated by at least 10 bp or so due to
// "mechanical" constraints. The embedding is always symmetric by default.
// NOTE: Ebp is the energy per embedding bp, typically -log(Pb).
void Landscapes::embed_landscape(Bkgr &bkgr, Sequence *seq, unsigned int ind) {
  
  if (Lembed.at(ind) == 0) return; // nothing to do if no embedding requested ..
  
  // Do embedding:
  if (en[ind].size() - 1 - 2*Lembed[ind] < 0) {
    std::cerr << "Seq is too short: impossible to embed the object in 2*"<<Lembed[ind]<<" extra bp!\n";
    exit(EXIT_FAILURE);
  }
  
  int Lsite = seq->Length() - en[ind].size() + 1;
  if (Lsite <= 0) {
    std::cerr << "Illegal site length in embed_landscape!\n";
    exit(EXIT_FAILURE);
  }
  
  std::vector<double> E_loc;
  std::vector<int> seqN_loc;
  
  for (unsigned int j=0; j<=(en[ind].size() - 1 - 2*Lembed[ind]); ++j) {
    seqN_loc.push_back(seqN[ind][j]);
    double Eembed = 0.0; // embedding bkgr energy
    for (unsigned int k=j; k<=(j+Lembed[ind]-1); ++k) {
      Eembed += bkgr.GetEb(seq->CharAt(k) );
    }
    for (unsigned int k=(j+Lembed[ind]+Lsite); k<=(j+2*Lembed[ind]+Lsite-1); ++k) {
      Eembed += bkgr.GetEb(seq->CharAt(k) );
    }
    E_loc.push_back(en[ind][j+Lembed[ind]] + Eembed);
  }
  
  // Reload arrays:
  seqN[ind].clear();
  en[ind].clear();
  for (unsigned int i=0; i<seqN_loc.size(); ++i) {
    seqN[ind].push_back(seqN_loc[i]);
  }
  for (unsigned int i=0; i<E_loc.size(); ++i) {
    en[ind].push_back(E_loc[i]);
  }
  
  printf("Embedded landscape %d in %d bp of bkgr sequence (%d bp on each side) ..\n",ind,2*Lembed[ind],Lembed[ind]);
}

// This function outputs the results (probs & occ).
void Landscapes::output_prob_occ(std::string const &out, std::string const &bkgr_choice, unsigned int Lseq, double Bw) {
  
  // Check array dims:
  for (unsigned int j=1; j<NLAND; ++j) { // over all landscapes except bkgr ..
    unsigned int prob_dim = prob[j-1].size();
    unsigned int occ_dim = occ[j-1].size();
    if (Lseq != prob_dim) {
      std::cerr << "Prob array mismatch: expected "<<Lseq<<" entries but have "<<prob_dim<<"!\n";
      exit(EXIT_FAILURE);
    }
    if (Lseq != occ_dim) {
      std::cerr << "Occ array mismatch: expected "<<Lseq<<" entries but have "<<occ_dim<<"!\n";
      exit(EXIT_FAILURE);
    }
  }
  
  // Output file:
  FILE *fp = fopen(out.c_str(),"w");
  if (fp == NULL) {
    fprintf(stderr,"Error writing to file: %s\n",out.c_str() );
    exit(EXIT_FAILURE);
  }
  
  fprintf(fp,"# Bw = %6.2f, bkgr_model = %s\n",Bw,bkgr_choice.c_str() );
  for (unsigned int j=1; j<NLAND; ++j) { // over all landscapes except bkgr ..
    fprintf(fp,"# <E(%d)> = %6.2f, var(E(%d)) = %6.2f\n",j,meanEs[j],j,varEs[j]);
  }
  // Header:
  fprintf(fp,"%8s","seqN");
  for (unsigned int j=1; j<NLAND; ++j) { // over all landscapes except bkgr ..
    fprintf(fp,"  %8s%d  %8s%d","Prob",j,"Occ",j);
  }
  fprintf(fp,"\n");
  // Results:
  for (unsigned int i=1; i<=Lseq; ++i) {
    fprintf(fp,"%8d",i);
    for (unsigned int j=1; j<NLAND; ++j) { // over all landscapes except bkgr ..
      fprintf(fp,"   %8.5f   %8.5f",prob[j-1][i-1],occ[j-1][i-1]);
    }
    fprintf(fp,"\n");
  }
  
  fclose(fp);
}

