/* sequence.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DYNAPRO implements a dynamic programming algorithm for predicting
DNA occupancies by arbitrary objects whose interaction energies with DNA are
known for each possible binding position.

The DYNAPRO method and its applications are described in the following
paper:

******************************************************************/

#include "sequence.h"
#include <math.h>

// This is Sequence class.

Sequence::Sequence(char *flname)
{
  FILE *fp = fopen(flname,"r");
  if (fp == NULL) {
    printf("Error reading file %s\n",flname);
    exit(1);
  }
  _seq = ReadFasta(fp);
  fclose(fp);

  _length = strlen(_seq);
  _indexseq = new char[_length];
  for (int i=0; i<_length; i++) {
    switch(_seq[i]) {
    case 'a':
    case 'A': _indexseq[i] = 0; break;
    case 'c':
    case 'C': _indexseq[i] = 1; break;
    case 'g': 
    case 'G': _indexseq[i] = 2; break;
    case 't':
    case 'T': _indexseq[i] = 3; break;
    default:  _indexseq[i] = 4;
    } 
  }

  _id = 0; // may be set from outside if relevant
  sprintf(_name,"No Name");
  _udata = NULL;
}

Sequence::Sequence(char *seq, int length, char *name)
{
  _seq = new char[length+1];
  strcpy(_seq,seq);
  _length = length;
  _indexseq = new char[_length];
  for (int i=0; i<_length; i++) {
    switch(_seq[i]) {
    case 'a':
    case 'A': _indexseq[i] = 0; break;
    case 'c':
    case 'C': _indexseq[i] = 1; break;
    case 'g': 
    case 'G': _indexseq[i] = 2; break;
    case 't':
    case 'T': _indexseq[i] = 3; break;
    default:  _indexseq[i] = 4;
    } 
  }

  _id = 0;
  strcpy(_name,name);
  _udata = NULL;
}

Sequence::Sequence(Sequence &src) {
  int length = src.Length();
  _seq = new char[length+1];
  _indexseq = new char[length];
  _length = length;
  for (int i=0; i<_length; i++) {
    _seq[i] = src._seq[i];
    _indexseq[i] = src._indexseq[i];
  }
  _seq[_length] = '\0';
  
  _id = 0;
  strcpy(_name,src._name);
  _udata = NULL;
}

Sequence::~Sequence()
{
  delete [] _seq;
  delete [] _indexseq;
}

char Sequence::CharAt(int index)
{
  return _seq[index];
}

void Sequence::Mask(int position)
{
  if (position < 0 || position >= _length) {
    printf("Error: masking of window requested at invalid position\n");
    exit(1);
  }
  _seq[position] = 'N';
  _indexseq[position] = 4;
  return;
}

char Sequence::IndexOfCharAt(int index)
{
  return _indexseq[index];
}

// This function returns an array of seq indices.
// NOTE: new memory is allocated inside!!
char *Sequence::SeqIndices(void) {
  char *ind = new char[_length];
  for (int j=0; j<_length; ++j) {
    ind[j] = _indexseq[j];
  }
  return ind;
}

// This function returns seq as a C-string.
// NOTE: new memory is allocated inside!!
char *Sequence::SeqString(void) {
  char *ss = new char[_length+1];
  for (int j=0; j<_length; ++j) {
    ss[j] = _seq[j];
  }
  ss[_length] = '\0';
  return ss;
}

bool Sequence::AmbiguousCharAt(int index)
{
  return (IndexOfCharAt(index) > 3);
}

int Sequence::Length()
{
  return _length;
}

char *Sequence::ReadFasta(FILE *fp)
{
  char line[MAX_LINE_LENGTH];
  do {
    fgets(line,MAX_LINE_LENGTH-1,fp);
  } while (line[0] == '#');
  if (line[0] != '>') {
    printf("Error: In Fasta file, each sequence must be preceded by a header line beginning with a '>'\n");
    exit(1);
  }
  char *sequence = new char[MAX_SEQ_LENGTH+1];
  sequence[0] = 0;
  int seq_pos = 0;
  while (fgets(line,MAX_LINE_LENGTH-1,fp)) {
    if (line[0]=='#') continue;
    if (strstr(line,">")) {
      strcpy(&(sequence[seq_pos]),"$");
      seq_pos ++;
      continue;   // if we want all sequences as one
    }

    // chomp
    int last_pos = strlen(line)-1;
    if (line[last_pos]=='\n' || line[last_pos]=='\r') line[last_pos] = 0;

    // concatenate
    strcpy(&(sequence[seq_pos]),line);
    seq_pos += strlen(line);
  }
  char *retval = new char[strlen(sequence)+1];
  strcpy(retval,sequence);
  delete [] sequence;

  return retval;
}

void Sequence::Print(int start, int stop)
{
  for (int i=start; i<=stop; i++) {
    printf("%c",_seq[i]);
  }
  printf("\n");
}

void Sequence::Name(char *name)
{
  strcpy(name,_name);
}

void Sequence::SetSpeciesIndex(int i)
{
  _id = i;
}

int Sequence::GetSpeciesIndex()
{
  return _id;
}

void Sequence::SetUserData(void *udata)
{
  _udata = udata;
}

void *Sequence::GetUserData()
{
  return _udata;
}

// This function mutates the seq at position specified by index.
// The new nucleotide is given by its nt_type - a number in the [0..3] range.
void Sequence::Mutate(int index, char nt_type) {
  assert(index>=0 && index<_length);
  assert(nt_type>=0 && nt_type<4);
  _indexseq[index] = nt_type;
  switch(nt_type)
    {
    case 0: _seq[index] = 'A'; break;
    case 1: _seq[index] = 'C'; break;
    case 2: _seq[index] = 'G'; break;
    case 3: _seq[index] = 'T'; break;
    default:  _seq[index] = 'N';
    }
}

